;; Generated package description from dts-mode.el  -*- no-byte-compile: t -*-
(define-package "dts-mode" "1.0.0.20221221.75311" "Major mode for Device Tree source files" '((emacs "24")) :commit "201d8e188fbe59c84e73d05097bb3eeb4495162e" :url "https://elpa.gnu.org/packages/dts-mode.html" :authors '(("Ben Gamari" . "ben@smart-cactus.org")) :maintainer '("Ben Gamari" . "ben@smart-cactus.org") :keywords '("languages"))
